<?php
/** @file notesviewconnector.php
 * Get data from a notes view.
 *
 * @author   Stefan Weber
 * @author   Visual Solutions
 * @version  1.0
 * @date     2006
 */


  /** @class ViewParser
   * Callback function class used to parse the domino xml data recieved from the notesserver.
   *
   * @dot
   *  digraph G {
   *    node [shape=record, fontname=Helvetica, fontsize=10];
   *    connectorclass [ label="class NotesViewConnector" URL="\ref NotesViewConnector"];
   *    parserclass [ label="class ViewParser" URL="\ref ViewParser"];
   *    connector [ label="{parseURL() | {<open>\<text\> | <data>string value | <close>\</text\>}}"];
   *    parser [ label="<start>StartElement() | <data>CharacterData() | <end>EndElement()"];
   *    {rank=same; "parserclass"; "parser";}
   *    connector:open -> parser:start [ arrowhead="open" ];
   *    connector:data -> parser:data [ arrowhead="open", label=" n*" ];
   *    connector:close -> parser:end [ arrowhead="open" ];
   *  }
   * @enddot
   * The class supports multiple calls of the xml_parse() function with fragmented XML data.
   *
   * @todo  Add a propper number and date formating when $preformat is FALSE.
   */
  class ViewParser
  {
    /// @privatesection
    var $preformat;         ///<  FALSE: a columnvalue is stored as a string, TRUE: a columnvalue is stored as an array of strings.

    var $tag = '';          ///<  Current XML tag, necessary to store the tagname while the CharacterData callback function does not pass it.
    var $toplevelentries;   ///<  0-indexed number of toplevel entries in the whole view.
    var $entryposition;     ///<  Current entryposition in the view. The entryposition in a hierarchical view can include subindexes like '3.5.1'.
    var $columnnumber;      ///<  Current columnnumber in the viewentry.
    var $listposition;      ///<  Current position in the itemlist, only used when $preformat is FALSE.

    var $result = array();  ///< Array to store the viewentries while parsing the XML data.


    /**
     * @public
     * Constuctor.
     *
     * @param   $preformat  int   If the parameter is FALSE each columnvalue is represented by a string, otherwise by an 0-indexed array of locale-neutral formated strings.
     */
    function ViewParser($preformat=false)
    {
      $this->preformat = $preformat;
    }

    /**
     * Callback function to handle the XML start tags.
     *
     * @param   $parser     object  Reference to the XML parser calling the handler.
     * @param   $tag        string  The name of the element for which this handler is called.
     * @param   $attribute  string  An associative array with the element's attributes (if any).
     */
    function StartElement($parser, $tag, $attribute)
    {
      $this->tag = $tag;  // store the currently parsed XML tag
      switch ($tag)
      {
      case 'viewentries': // top level XML tag
        $this->toplevelentries = $attribute['toplevelentries'];
        break;

      case 'viewentry':   // start the next viewentry
        $this->entryposition = $attribute['position'];
        if ( isset($attribute['unid']) )
          $this->result[$this->entryposition]['unid'] = $attribute['unid'];
        if ( isset($attribute['categorytotal']) )
          $this->result[$this->entryposition]['categorytotal'] = $attribute['categorytotal'];
        break;

      case 'entrydata':   // start the next column
        $this->columnnumber = $attribute['columnnumber'];
        if ( isset($attribute['category']) )
          $this->result[$this->entryposition]['category'] = $attribute['category'];
        $this->listposition = 0;  // only used when $preformat is FALSE
        break;

      case 'number':      // Todo: propper formating
      case 'datetime':
      case 'text':        // initialze the column data to support multiple xml_parse() function calls
        if ( $this->preformat )
          $this->result[$this->entryposition]['column'][$this->columnnumber] = '';
        else
          $this->result[$this->entryposition]['column'][$this->columnnumber][$this->listposition] = '';
      }
    }

    /**
     * Callback function to handle the XML end tags.
     *
     * @param   $parser  object  Reference to the XML parser calling the handler.
     * @param   $tag     string  The name of the element for which this handler is called.
     */
    function EndElement($parser, $tag)
    {
      switch ($tag)
      {
      case 'number':      // Todo: propper formating
                          // this is not really useful
//        if ( isset($this->result[$this->entryposition]['column'][$this->columnnumber][$this->listposition]) )
//          $this->result[$this->entryposition]['column'][$this->columnnumber][$this->listposition] .= ' Euro';
      case 'datetime':
      case 'text':
        ++$this->listposition;  // only used when $preformat is FALSE
      }
      $this->tag = '';  // no XML tag is currently parsed
    }

    /**
     * Callback function to handle the entry character data.
     *
     * @param   $parser  object  Reference to the XML parser calling the handler.
     * @param   $cdata   string  Character data as a string (character data of one element may be split into several function calls).
     */
    function CharacterData($parser, $cdata)
    {
      switch ($this->tag)
      {
      case 'number':
      case 'datetime':
      case 'text':      // concatenate the $cdata fragments to support multiple xml_parse() function calls
        if ( $this->preformat )
          $this->result[$this->entryposition]['column'][$this->columnnumber] .= $cdata;
        else
          $this->result[$this->entryposition]['column'][$this->columnnumber][$this->listposition] .= $cdata;
      }
    }

    /**
     * @public
     * Get the result array after parsing the XML data.
     */
    function GetData()
    {
    	return $this->result;
    }
  }


  /** @class NotesViewConnector
   * Class to handle the communication with a notesserver getting data from a notes view.
   *
   * Usage similar to a mysql database connection.
   * See http://www.visol.de/php for details.
   *
   * @todo  Add support for hierarchical views.
   */
  class NotesViewConnector
  {
    /// @privatesection
    var $preformat;        ///<  Used to set the notesview URL parametern and passed through to ViewParser class.
    var $viewurl = '';     ///<  Basic part of the notesview URL, completed with Start= and Count= URL parameters.

    var $buffer = null;    ///<  Array to cache some viewentries to get a better performance.
    var $bufferstart = 0;  ///<  First entrynumber in buffer.
    var $buffersize = 3;   ///<  Buffersize must not exceed the "Maximum lines per view page" setting in domino directory.
    var $cursor = 0;       ///<  Current viewentry returend by getViewEntry function.


    /**
     * @public
     * Creates a connection class associated to the specified notes view.
     *
     * The constructor just instantiates the PHP class, the notesserver is not connected.
     *
     * @param   $servername    string   DNS name or IP adrress of the notesserver, may include a port number. Example: 'www.visol.de:80'
     * @param   $databasename  string   Name or databaseid and path of the target notesdatabase. Example: 'php/ezdemo.nsf'
     * @param   $viewname      string   Name or aliasname of the notesview. Example: 'menuflat'
     * @param   $ppreformat    int      TRUE causes all data types to be converted to text on the notes server (numbers, dates and all type of lists). @n
     *                                  Caution! The server's locale is used for all formatting. @n
     *                                  If this argument is FALSE, each column is represented by an 0-indexed array of locale-neutral formated strings.
     */
    function NotesViewConnector( $servername, $databasename, $viewname, $ppreformat=false )
    {
      $this->viewurl = 'http://'.$servername.'/'.$databasename.'/'.$viewname.'?ReadViewEntries&ExpandView';
      if ($ppreformat)
        $this->viewurl .= '&Preformat';
      $this->preformat = $ppreformat;
      $this->buffer = null;
      $this->bufferstart = 0;
      $this->cursor = 0;
    }

    /**
     * @public
     * Returns an array representing a viewentry in a notesview.
     *
     * Returns a array with column values indexed by the column position
     * representing the next or requested viewentry in the associated notesview. @n
     * Returns FALSE if there are no viewentries left in the notesview,
     * or the requested viewentry does not exist.
     *
     * If the viewentry represents a notesdocument an additional item contains the unique id of that document.
     *
     * @codeexample
     * Array( [docunid] => '8807DDFF375EA5C7C125713900464C91',
     *        [column] => Array( [0]=>'Mensa', [1]=>'04/06/2006', [2]=>'Maccaroni', [3]=>'Vegetarisch,Wahlessen 3', [4]=>'2.00' )
     *      )
     * @endcode
     *
     * @warning The column position indexes in the returned array are not necessarily continuous and
     *          not all viewentries must have the same columns, especially not in hierarchical views.
     *
     * @param   $entrynumber  string  viewentry position in the notesview. If the position is omitted the next viewentry is retrived.
     *
     * @return  array         Representing a single viewentry or FALSE when there is no viewentry.
     */
    function getViewEntry( $entrynumber=0 )
    {
      if ( $entrynumber!=0 )
        $this->cursor = $entrynumber;
      else
        ++$this->cursor;

      if ( $this->bufferstart==0 || $this->cursor<$this->bufferstart || $this->cursor>=$this->bufferstart+$this->buffersize )
      {
        $this->buffer = $this->parseURL($this->viewurl.'&Start='.$this->cursor.'&Count='.$this->buffersize);
        $this->bufferstart = $this->cursor;
      }
      if ( isset($this->buffer[$this->cursor]) )
        return $this->buffer[$this->cursor];
      else
        return false;
    }

    /**
     * @public
     * Main function to get the domino view data.
     *
     * Domino XML data provided by a notes server will by parsed and stored in an array.
     *
     * @param   $url  string  Propper URL string to get XML view data from a notes server.
     */
    function parseURL( $url )
    {
      $fp = fopen( $url, 'r' );

      $parserclass = new ViewParser($this->preformat);
      $parser = xml_parser_create();
      xml_set_object($parser, $parserclass);
      xml_set_element_handler($parser, 'StartElement', 'EndElement');
      xml_set_character_data_handler($parser, 'CharacterData');
      xml_parser_set_option($parser, XML_OPTION_CASE_FOLDING, 0);
  //    xml_parser_set_option($parser, XML_OPTION_SKIP_WHITE,  1);  // should not be necessary here

      while ( $buf = fread($fp, 1024) )
        xml_parse($parser, $buf, feof($fp));

      xml_parser_free($parser);

      fclose($fp);
      return $parserclass->GetData();
    }

  }

?>